@@ -70,7 +70,7 @@
 
 // Even minor versions (e.g. 1.0 or 1.2) are stable versions.
 // Odd minor versions (e.g. 1.1 or 1.3) are development versions.
-#define PROGRAM_VERSION "1.13"
+#define PROGRAM_VERSION "1.16"
 
 namespace {
 
@@ -240,6 +240,7 @@ enum {
   KEY_NO_SPECIALS,
   KEY_NO_SYMLINKS,
   KEY_NO_HARDLINKS,
+  KEY_NO_XATTRS,
   KEY_DEFAULT_PERMISSIONS,
 #if FUSE_USE_VERSION >= 30
   KEY_DIRECT_IO,
@@ -272,6 +273,7 @@ fuse_opt const g_fuse_opts[] = {
     FUSE_OPT_KEY("nospecials", KEY_NO_SPECIALS),
     FUSE_OPT_KEY("nosymlinks", KEY_NO_SYMLINKS),
     FUSE_OPT_KEY("nohardlinks", KEY_NO_HARDLINKS),
+    FUSE_OPT_KEY("noxattrs", KEY_NO_XATTRS),
     FUSE_OPT_KEY("default_permissions", KEY_DEFAULT_PERMISSIONS),
 #if FUSE_USE_VERSION >= 30
     FUSE_OPT_KEY("direct_io", KEY_DIRECT_IO),
@@ -289,6 +291,7 @@ bool g_force = false;
 bool g_specials = true;
 bool g_symlinks = true;
 bool g_hardlinks = true;
+bool g_xattrs = true;
 bool g_default_permissions = false;
 #if FUSE_USE_VERSION >= 30
 bool g_direct_io = false;
@@ -404,6 +407,75 @@ gid_t const g_gid = getgid();
 using Clock = std::chrono::system_clock;
 time_t const g_now = Clock::to_time_t(Clock::now());
 
+// A string view and its hashed value.
+struct HashedStringView {
+  std::string_view string;
+  size_t hashed_value = std::hash<std::string_view>()(string);
+  explicit HashedStringView(std::string_view s) : string(s) {}
+};
+
+// A string and its hashed value.
+struct HashedString {
+  std::string string;
+  size_t hashed_value;
+  HashedString(const HashedStringView& x)
+      : string(x.string), hashed_value(x.hashed_value) {}
+};
+
+// Function object that compares HashedStringView and HashedString objects for
+// equality.
+struct IsEqual {
+  using is_transparent = std::true_type;
+  bool operator()(const auto& a, const auto& b) const {
+    return a.hashed_value == b.hashed_value && a.string == b.string;
+  }
+};
+
+// Function object that gets the hashed value of HashedStringView and
+// HashedString objects.
+struct Hash {
+  using is_transparent = std::true_type;
+  size_t operator()(const auto& x) const { return x.hashed_value; }
+};
+
+// Set of unique (i.e. deduplicated) strings. Lazily constructed and populated.
+// Never destructed.
+using HashedStrings = std::unordered_set<HashedString, Hash, IsEqual>;
+HashedStrings* g_unique_strings = nullptr;
+
+// Gets a pointer to the unique registered string matching the given string, or
+// null if no such string has been registered.
+const HashedString* GetUniqueOrNull(std::string_view const s) {
+  if (!g_unique_strings) {
+    return nullptr;
+  }
+
+  HashedStrings::const_iterator const it =
+      g_unique_strings->find(HashedStringView(s));
+  if (it == g_unique_strings->cend()) {
+    return nullptr;
+  }
+
+  assert(it->string == s);
+  assert(it->string.data() != s.data());
+  return &*it;
+}
+
+// Gets a non-null pointer to the unique registered string matching the given
+// string, creating and registering it if necessary.
+const HashedString* GetOrCreateUnique(std::string_view const s) {
+  if (!g_unique_strings) {
+    g_unique_strings = new HashedStrings();
+  }
+
+  HashedStrings::const_iterator const it =
+      g_unique_strings->insert(HashedStringView(s)).first;
+  assert(it != g_unique_strings->cend());
+  assert(it->string == s);
+  assert(it->string.data() != s.data());
+  return &*it;
+}
+
 // Converts a string to ASCII lower case.
 std::string ToLower(std::string_view const s) {
   std::string r(s);
@@ -1193,8 +1265,11 @@ struct Reader : bi::list_base_hook<LinkMode> {
         {"brotli", SET_FILTER_COMMAND(brotli)},
         {"bz2", SET_FILTER(BZIP2)},
         {"bzip2", SET_FILTER(BZIP2)},
-        {"grz", SET_FILTER(GRZIP)},
-        {"grzip", SET_FILTER(GRZIP)},
+        // Work around https://github.com/libarchive/libarchive/issues/2513
+        // {"grz", SET_FILTER(GRZIP)},
+        // {"grzip", SET_FILTER(GRZIP)},
+        {"grz", SET_FILTER_COMMAND(grzip)},
+        {"grzip", SET_FILTER_COMMAND(grzip)},
         {"gz", SET_FILTER(GZIP)},
         {"gzip", SET_FILTER(GZIP)},
         {"lrz", SET_FILTER(LRZIP)},
@@ -1205,6 +1280,7 @@ struct Reader : bi::list_base_hook<LinkMode> {
         {"lzma", SET_FILTER(LZMA)},
         // Work around https://github.com/libarchive/libarchive/issues/2513
         // {"lzo", SET_FILTER(LZOP)},
+        // {"lzop", SET_FILTER(LZOP)},
         {"lzo", SET_FILTER_COMMAND(lzop)},
         {"lzop", SET_FILTER_COMMAND(lzop)},
         {"uu", SET_FILTER(UU)},
@@ -1543,6 +1619,14 @@ struct Node {
   dev_t rdev = 0;
   i64 nlink = 1;
 
+  // Extended attributes.
+  struct Attribute {
+    const HashedString* key;
+    std::string value;
+  };
+
+  std::vector<Attribute> attributes;
+
   // Number of entries whose name have initially collided with this file node.
   int collision_count = 0;
 
@@ -2161,8 +2245,9 @@ void ProcessEntry(Reader& r) {
     // Entry is a hard link.
     if (g_hardlinks) {
       // Save it for further resolution.
-      g_hardlinks_to_resolve.emplace_back(i, std::move(path),
-                                          Path(s).Normalized());
+      g_hardlinks_to_resolve.push_back({.index_within_archive = i,
+                                        .source_path = std::move(path),
+                                        .target_path = Path(s).Normalized()});
     } else {
       LOG(DEBUG) << "Skipped hard link "
                  << " [" << i << "] " << Path(path) << " -> " << Path(s);
@@ -2260,6 +2345,23 @@ void ProcessEntry(Reader& r) {
     node->size = r.GetEntrySize();
   }
 
+  // Extract extended attributes.
+  if (g_xattrs) {
+    if (int const n = archive_entry_xattr_reset(e)) {
+      node->attributes.reserve(n);
+      const char* key;
+      const void* value;
+      size_t len;
+      while (archive_entry_xattr_next(e, &key, &value, &len) == ARCHIVE_OK) {
+        assert(key);
+        assert(value);
+        node->attributes.push_back(
+            {.key = GetOrCreateUnique(key),
+             .value = std::string(static_cast<const char*>(value), len)});
+      }
+    }
+  }
+
   // Check password if necessary.
   r.CheckPassword();
 
@@ -2470,7 +2572,7 @@ int GetAttr(const char* const path,
       return -ENOENT;
     }
   }
-  
+
   const Node* const t = n->GetTarget();
   assert(t);
 
@@ -2479,6 +2581,123 @@ int GetAttr(const char* const path,
   return 0;
 }
 
+int GetXattr(const char* const path,
+             const char* const xattr_name,
+             char* const dst_ptr,
+#ifdef __APPLE__
+             size_t const dst_len,
+             uint32_t) {
+#else
+             size_t const dst_len) {
+#endif
+  assert(path);
+  assert(xattr_name);
+
+  const Node* const node = FindNode(path);
+  if (!node) {
+    LOG(ERROR) << "Cannot get xattr " << std::quoted(xattr_name) << " of "
+               << Path(path) << ": No such file or directory";
+    return -ENOENT;
+  }
+
+  if (node->attributes.empty()) {
+    // The node has no extended attributes.
+    LOG(DEBUG) << *node << " has no xattr " << std::quoted(xattr_name);
+    return -ENODATA;
+  }
+
+  const HashedString* const key = GetUniqueOrNull(xattr_name);
+  if (!key) {
+    // No extended attribute has ever been recorded with the given name.
+    LOG(DEBUG) << *node << " has no xattr " << std::quoted(xattr_name);
+    return -ENODATA;
+  }
+
+  auto const it =
+      std::ranges::find(node->attributes, key, &Node::Attribute::key);
+  if (it == node->attributes.end()) {
+    // The node has some extended attributes, but none matching the given name.
+    LOG(DEBUG) << *node << " has no xattr " << std::quoted(xattr_name);
+    return -ENODATA;
+  }
+
+  assert(it->key->string == xattr_name);
+  const std::string& value = it->value;
+
+  if (dst_len > 0) {
+    assert(dst_ptr);
+    if (dst_len < value.size()) {
+      LOG(ERROR) << "Cannot get xattr " << std::quoted(xattr_name) << " of "
+                 << *node << ": The destination buffer of " << dst_len
+                 << " bytes is too small: Needs at least " << value.size()
+                 << " bytes";
+      return -ERANGE;
+    }
+
+    std::copy_n(value.data(), value.size(), dst_ptr);
+  }
+
+  if (value.size() > std::numeric_limits<int>::max()) {
+    LOG(ERROR) << "Cannot get xattr " << std::quoted(xattr_name) << " of "
+               << *node << ": The value is " << value.size()
+               << " bytes long, which is greater than MAX_INT";
+    return -E2BIG;
+  }
+
+  LOG(DEBUG) << "Get xattr " << std::quoted(xattr_name) << " of " << *node
+             << " -> " << value.size() << " bytes";
+  return static_cast<int>(value.size());
+}
+
+int ListXattr(const char* const path,
+              char* const dst_ptr,
+              size_t const dst_len) {
+  const Node* const node = FindNode(path);
+  if (!node) {
+    LOG(ERROR) << "Cannot list xattrs of " << Path(path)
+               << ": No such file or directory";
+    return -ENOENT;
+  }
+
+  size_t total_bytes = 0;
+  if (dst_len == 0) {
+    // Compute required buffer size.
+    for (const Node::Attribute& a : node->attributes) {
+      const std::string& key = a.key->string;
+      total_bytes += key.size() + 1;
+    }
+  } else {
+    assert(dst_ptr);
+    std::span<char> dst(dst_ptr, dst_len);
+    for (const Node::Attribute& a : node->attributes) {
+      const std::string& key = a.key->string;
+      const size_t n = key.size() + 1;
+      if (dst.size() < n) {
+        LOG(ERROR) << "Cannot list " << node->attributes.size() << " xattrs of "
+                   << *node << ": The destination buffer of " << dst_len
+                   << " bytes is too small";
+        return -ERANGE;
+      }
+
+      // Copy the NUL terminator as well.
+      std::copy_n(key.c_str(), n, dst.data());
+      dst = dst.subspan(n);
+      total_bytes += n;
+    }
+  }
+
+  if (total_bytes > std::numeric_limits<int>::max()) {
+    LOG(ERROR) << "Cannot list " << node->attributes.size() << " xattrs of "
+               << *node << ": The list is " << total_bytes
+               << " bytes long, which is greater than MAX_INT";
+    return -E2BIG;
+  }
+
+  LOG(DEBUG) << "List " << node->attributes.size() << " xattrs of " << *node
+             << " -> " << total_bytes << " bytes";
+  return static_cast<int>(total_bytes);
+}
+
 int ReadLink(const char* const path, char* const buf, size_t const size) {
   assert(path);
   assert(buf);
@@ -2781,6 +3000,8 @@ fuse_operations const operations = {
   .read = Read,
   .statfs = StatFs,
   .release = Release,
+  .getxattr = GetXattr,
+  .listxattr = ListXattr,
   .opendir = OpenDir,
   .readdir = ReadDir,
 #if FUSE_USE_VERSION >= 30
@@ -2858,6 +3079,10 @@ int ProcessArg(void*, const char* const arg, int const key, fuse_args*) {
       g_hardlinks = false;
       return DISCARD;
 
+    case KEY_NO_XATTRS:
+      g_xattrs = false;
+      return DISCARD;
+
     case KEY_DEFAULT_PERMISSIONS:
       g_default_permissions = true;
       return KEEP;
@@ -2924,7 +3149,7 @@ class NumPunct : public std::numpunct<char> {
 };
 
 void PrintUsage() {
-  std::cerr << "usage: " PROGRAM_NAME
+  std::cout << "usage: " PROGRAM_NAME
                R"( [options] <archive_file> [mount_point]
 
 general options:
@@ -2942,13 +3167,14 @@ general options:
     -o nospecials          no special files (FIFOs, sockets, devices)
     -o nosymlinks          no symlinks
     -o nohardlinks         no hard links
+    -o noxattrs            no extended attributes
     -o dmask=M             directory permission mask in octal (default 0022)
     -o fmask=M             file permission mask in octal (default 0022))"
 #if FUSE_USE_VERSION >= 30
                R"(
     -o direct_io           use direct I/O)"
 #endif
-               "\n\n";
+      "\n\n" << std::flush;
 }
 
 }  // namespace
@@ -2993,23 +3219,24 @@ int main(int const argc, char** const argv) try {
   }
 
   if (g_version) {
-    std::cerr << PROGRAM_NAME " version: " PROGRAM_VERSION "\n";
-    std::cerr << "libarchive version: " << archive_version_string() << "\n";
+    std::cout << PROGRAM_NAME " version: " PROGRAM_VERSION "\n";
+    std::cout << "libarchive version: " << archive_version_string() << "\n";
     if (const char* const s = archive_bzlib_version()) {
-      std::cerr << "bzlib version: " << s << "\n";
+      std::cout << "bzlib version: " << s << "\n";
     }
     if (const char* const s = archive_liblz4_version()) {
-      std::cerr << "liblz4 version: " << s << "\n";
+      std::cout << "liblz4 version: " << s << "\n";
     }
     if (const char* const s = archive_liblzma_version()) {
-      std::cerr << "liblzma version: " << s << "\n";
+      std::cout << "liblzma version: " << s << "\n";
     }
     if (const char* const s = archive_libzstd_version()) {
-      std::cerr << "libzstd version: " << s << "\n";
+      std::cout << "libzstd version: " << s << "\n";
     }
     if (const char* const s = archive_zlib_version()) {
-      std::cerr << "zlib version: " << s << "\n";
+      std::cout << "zlib version: " << s << "\n";
     }
+    std::cout.flush();
 
     fuse_opt_add_arg(&args, "--version");
     fuse_main(args.argc, args.argv, &operations, nullptr);
@@ -3043,11 +3270,11 @@ int main(int const argc, char** const argv) try {
   int mount_point_parent_fd =
       open(!mount_point_parent.empty() ? mount_point_parent.c_str() : ".",
 #if defined(O_PATH)
-           O_DIRECTORY | O_PATH); // Linux, FreeBSD >= 13
+           O_DIRECTORY | O_PATH);  // Linux, FreeBSD >= 13
 #elif defined(O_EXEC)
-           O_DIRECTORY | O_EXEC); // macOS, FreeBSD <= 12
+           O_DIRECTORY | O_EXEC);  // FreeBSD <= 12
 #else
-           O_DIRECTORY | O_RDONLY); // OpenBSD
+           O_DIRECTORY | O_RDONLY);  // OpenBSD, macOS
 #endif
   if (mount_point_parent_fd < 0) {
     PLOG(ERROR) << "Cannot access directory " << Path(mount_point_parent);
