@@ -70,6 +70,19 @@ def GetTree(root, use_md5=True):
             for entry in os.scandir(path):
                 scan(entry.path, entry.stat(follow_symlinks=False))
 
+        try:
+            attrs = os.listxattr(path)
+            xattr_dict = dict()
+            for attr in attrs:
+                try:
+                    value = os.getxattr(path, attr)
+                    xattr_dict[attr] = value.decode('utf-8')
+                except OSError as e:
+                    xattr_dict[attr] = f'<error: {e.errno}>'
+            line['xattr'] = xattr_dict
+        except OSError as e:
+            line['xattr'] = e.errno
+
     st = os.stat(root, follow_symlinks=False)
 
     # On some systems, the mount point is not immediately functional.
@@ -251,14 +264,16 @@ def TestArchiveWithOptions(options=[]):
 
     for zip_name in [
             'archive.7z', 'archive.rar', 'archive.tar', 'archive.tar.b64',
-            'archive.tar.br', 'archive.tar.bz2', 'archive.tar.gz',
-            'archive.tar.lrz', 'archive.tar.lz', 'archive.tar.lz4',
-            'archive.tar.lzma', 'archive.tar.lzo', 'archive.tar.uu',
-            'archive.tar.xz', 'archive.tar.Z', 'archive.tar.zst',
-            'archive.taz', 'archive.tb2', 'archive.tbz', 'archive.tbz2',
-            'archive.tgz', 'archive.tlz', 'archive.tlz4', 'archive.tlzma',
-            'archive.txz', 'archive.tz', 'archive.tz2', 'archive.tzst',
-            'archive.zip', 'compressed.tar'
+            'archive.tar.br', 'archive.tbr', 'archive.tar.bz2',
+            'archive.tar.gz', 'archive.tar.lrz', 'archive.tar.lz',
+            'archive.tar.lz4', 'archive.tar.lzma', 'archive.tar.lzo',
+            'archive.tar.uu', 'archive.tar.xz', 'archive.tar.Z',
+            'archive.tar.zst', 'archive.taz', 'archive.tb2', 'archive.tbz',
+            'archive.tbz2', 'archive.tgz', 'archive.tlz', 'archive.tlzip',
+            'archive.tlz4', 'archive.tlzma', 'archive.txz', 'archive.tz',
+            'archive.tz2', 'archive.tzs', 'archive.tzst', 'archive.tzstd',
+            'archive.zip', 'compressed.tar', 'lz_is_lzip.tlz',
+            'lz_is_lzma.tlz',
     ]:
         MountArchiveAndCheckTree(zip_name, want_tree, options=options)
 
@@ -1285,25 +1300,94 @@ def TestInvalidArchive():
     # https://github.com/google/fuse-archive/issues/36
     CheckArchiveMountingError('truncated.7z', 32)
 
+    # 7Z encryption is not supported
+    CheckArchiveMountingError("encrypted.7z", 22, password='password')
+    CheckArchiveMountingError("encrypted-solidly.7z", 22, password='password')
+
     if os.getuid() != 0:
         with tempfile.NamedTemporaryFile() as f:
             os.chmod(f.name, 0)
             CheckArchiveMountingError(f.name, 11)
 
+# Tests extended attributes functionality
+def TestExtendedAttributes(options=[]):
+    # Test normal xattr archive
+    zip_name = 'many-xattrs.tar.gz'
+    want_tree = {
+        ".": {"ino": 1, "mode": "drwxr-xr-x", "nlink": 2},
+        "file.txt": {
+            "xattr": {
+                "user.attr_{:04d}".format(i): f"value_{i}" for i in range(1, 11)
+            },
+        },
+    }
+    MountArchiveAndCheckTree(zip_name, want_tree, options=options, use_md5=False)
+
+    want_tree = {
+        ".": {"ino": 1, "mode": "drwxr-xr-x", "nlink": 2},
+        "file.txt": {"xattr": {}},
+    }
+    MountArchiveAndCheckTree(zip_name, want_tree, options=options + ['-o', 'noxattrs'], use_md5=False)
+
+    # Test very long name xattr cases
+    zip_name = 'long-xattr-name.tar.gz'
+    want_tree = {
+        ".": {"ino": 1, "mode": "drwxr-xr-x", "nlink": 2},
+        "file.txt": {
+            # "xattr": {
+            #     'user.' + 'a' * 200 + '_!@#$%25^&*()': 'Special characters in long name',
+            #     'user.' + 'x' * 245: 'This is a very long attribute name',
+            # },
+        },
+    }
+    MountArchiveAndCheckTree(zip_name, want_tree, options=options, use_md5=False)
+
+    want_tree = {
+        ".": {"ino": 1, "mode": "drwxr-xr-x", "nlink": 2},
+        "file.txt": {"xattr": {}},
+    }
+    MountArchiveAndCheckTree(zip_name, want_tree, options=options + ['-o', 'noxattrs'], use_md5=False)
+
+    # Test xattr error cases
+    zip_name = 'xattr-errors.tar.gz'
+    want_tree = {
+        ".": {"ino": 1, "mode": "drwxr-xr-x", "nlink": 2},
+        "file.txt": {
+            "xattr": {
+                'user.normal_attr': 'valid value',
+                'user.corrupt_attr': '-',
+            },
+        },
+    }
+    MountArchiveAndCheckTree(zip_name, want_tree, options=options, use_md5=False)
+
+    want_tree = {
+        ".": {"ino": 1, "mode": "drwxr-xr-x", "nlink": 2},
+        "file.txt": {"xattr": {}},
+    }
+    MountArchiveAndCheckTree(zip_name, want_tree, options=options + ['-o', 'noxattrs'], use_md5=False)
+
 
 logging.getLogger().setLevel('INFO')
 
 TestArchiveWithOptions()
 TestArchiveWithOptions(['-o', 'nocache'])
+TestArchiveWithOptions(['-o', 'lazycache'])
 TestHardlinks()
 TestHardlinks(['-o', 'nocache'])
+TestHardlinks(['-o', 'lazycache'])
 TestArchiveWithSpecialFiles()
 TestEncryptedArchive()
 TestEncryptedArchive(['-o', 'nocache'])
+TestEncryptedArchive(['-o', 'lazycache'])
+TestExtendedAttributes()
+TestExtendedAttributes(['-o', 'nocache'])
+TestExtendedAttributes(['-o', 'lazycache'])
 TestInvalidArchive()
 TestMasks()
 TestArchiveWithManyFiles()
 TestBigArchiveRandomOrder(['-o', 'direct_io'])
+TestBigArchiveRandomOrder(['-o', 'lazycache,direct_io'])
 TestBigArchiveStreamed(['-o', 'nocache,direct_io'])
 
 if error_count:
